@@ -97,9 +97,10 @@ use crate::utils::update_data;
 use crate::utils::{compute_cstr_scales, compute_obj_scale, compute_wb2s_scale};
 use crate::utils::{ei, wb2s};
 use egobox_doe::{Lhs, LhsKind, SamplingMethod};
-use egobox_moe::{CorrelationSpec, Moe, MoeFit, MoePredict, RegressionSpec};
+use egobox_moe::{CorrelationSpec, Moe, MoeParams, RegressionSpec, Surrogate};
 use env_logger::{Builder, Env};
 use finitediff::FiniteDiff;
+use linfa::{traits::Fit, Dataset, ParamGuard};
 use log::{debug, info};
 use ndarray::{concatenate, s, Array, Array1, Array2, ArrayBase, Axis, Data, Ix1, Ix2, Zip};
 use ndarray_linalg::Scalar;
@@ -112,22 +113,19 @@ use rand_isaac::Isaac64Rng;
 const DOE_INITIAL_FILE: &str = "egor_initial_doe.npy";
 const DOE_FILE: &str = "egobox_doe.npy";
 
-#[derive(Clone, Copy, Debug)]
-pub struct ApproxValue {
-    pub value: f64,
-    pub tolerance: f64,
-}
-
 /// Data used by internal infill criteria to be optimized using NlOpt
-pub struct ObjData<F> {
+struct ObjData<F> {
     pub scale_obj: F,
     pub scale_cstr: Array1<F>,
     pub scale_wb2: F,
 }
 
-/// An interface for "function under optimization" evaluation
-pub trait Evaluator {
-    fn eval(&self, x: &Array2<f64>) -> Array2<f64>;
+impl<R: Rng + SeedableRng + Clone> SurrogateBuilder for MoeParams<f64, R> {
+    fn train(&self, xt: &Array2<f64>, yt: &Array2<f64>) -> Result<Box<dyn Surrogate>> {
+        let checked = self.check_ref()?;
+        let moe = checked.fit(&Dataset::new(xt.to_owned(), yt.to_owned()))?;
+        Ok(moe).map(|moe| Box::new(moe) as Box<dyn Surrogate>)
+    }
 }
 
 /// EGO optimization parameterization
@@ -139,7 +137,7 @@ pub struct Egor<'a, O: GroupFunc, R: Rng> {
     /// Number of starts for multistart approach used for hyperparameters optimization
     pub n_start: usize,
     /// Number of parallel points evaluated for each "function evaluation"
-    pub n_parallel: usize,
+    pub q_parallel: usize,
     /// Number of initial doe drawn using Latin hypercube sampling
     /// Note: n_doe > 0; otherwise n_doe = max(xdim+1, 5)
     pub n_doe: usize,
@@ -153,7 +151,7 @@ pub struct Egor<'a, O: GroupFunc, R: Rng> {
     pub doe: Option<Array2<f64>>,
     /// Matrix (nx, 2) of [lower bound, upper bound] of the nx components of x
     pub xlimits: Array2<f64>,
-    /// Parallel strategy used to define several points (n_parallel) evaluations at each iteration
+    /// Parallel strategy used to define several points (q_parallel) evaluations at each iteration
     pub q_ei: QEiStrategy,
     /// Criterium to select next point to evaluate
     pub infill: InfillStrategy,
@@ -173,11 +171,14 @@ pub struct Egor<'a, O: GroupFunc, R: Rng> {
     pub outdir: Option<String>,
     /// If true use <outdir> to retrieve and start from previous results
     pub hot_start: bool,
-    /// MoE parameters (see [egobox_moe])
+    /// An optional surrogate builder used to model objective and constraint
+    /// functions, otherwise [mixture of expert](egobox_moe) is used
     /// Note: if specified takes precedence over individual settings
-    pub moe_params: Option<&'a dyn MoeFit>,
-    /// An evaluator used to run the function under optimization
-    pub evaluator: Option<&'a dyn Evaluator>,
+    pub surrogate_builder: Option<&'a dyn SurrogateBuilder>,
+    /// An optional pre-processor to preprocess continuous input given
+    /// to the function under optimization, specially used with mixed integer
+    /// function optimization
+    pub pre_proc: Option<&'a dyn PreProcessor>,
     /// The function under optimization f(x) = [objective, cstr1, ..., cstrn], (n_cstr+1 size)
     pub obj: O,
     /// A random generator used to get reproductible results.
@@ -186,12 +187,21 @@ pub struct Egor<'a, O: GroupFunc, R: Rng> {
 }
 
 impl<'a, O: GroupFunc> Egor<'a, O, Isaac64Rng> {
+    /// Constructor of the optimization of the function `f`
+    ///
+    /// The function `f` shoud return an objective but also constraint values if any.
+    /// Design space is specified by the 2D array `xlimits` which is `[nx, 2]`-shaped and
+    /// constains lower and upper bounds of `x` components.
     pub fn new(f: O, xlimits: &ArrayBase<impl Data<Elem = f64>, Ix2>) -> Egor<'a, O, Isaac64Rng> {
         Self::new_with_rng(f, xlimits, Isaac64Rng::from_entropy())
     }
 }
 
 impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
+    /// Constructor of the optimization of the function `f` with specified random generator
+    /// to get reproducibility.
+    ///
+    /// See [`Egor::new()`]
     pub fn new_with_rng(f: O, xlimits: &ArrayBase<impl Data<Elem = f64>, Ix2>, rng: R) -> Self {
         let env = Env::new().filter_or("EGOBOX_LOG", "info");
         let mut builder = Builder::from_env(env);
@@ -200,7 +210,7 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         Egor {
             n_eval: 20,
             n_start: 20,
-            n_parallel: 1,
+            q_parallel: 1,
             n_doe: 0,
             n_cstr: 0,
             cstr_tol: 1e-6,
@@ -216,113 +226,151 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
             expected: None,
             outdir: None,
             hot_start: false,
-            moe_params: None,
-            evaluator: None,
+            surrogate_builder: None,
+            pre_proc: None,
             obj: f,
             rng,
         }
     }
 
+    /// Sets allowed number of evaluation of the function under optimization
     pub fn n_eval(&mut self, n_eval: usize) -> &mut Self {
         self.n_eval = n_eval;
         self
     }
 
+    /// Sets the number of runs of infill strategy optimizations (best result taken)
     pub fn n_start(&mut self, n_start: usize) -> &mut Self {
         self.n_start = n_start;
         self
     }
 
-    pub fn n_parallel(&mut self, n_parallel: usize) -> &mut Self {
-        self.n_parallel = n_parallel;
+    /// Sets Number of parallel evaluations of the function under optimization
+    pub fn q_parallel(&mut self, q_parallel: usize) -> &mut Self {
+        self.q_parallel = q_parallel;
         self
     }
 
+    /// Number of samples of initial LHS sampling (used when DOE not provided by the user)
+    ///
+    /// When 0 a number of points is computed automatically regarding the number of input variables
+    /// of the function under optimization.
     pub fn n_doe(&mut self, n_doe: usize) -> &mut Self {
         self.n_doe = n_doe;
         self
     }
 
+    /// Sets the number of constraint functions
     pub fn n_cstr(&mut self, n_cstr: usize) -> &mut Self {
         self.n_cstr = n_cstr;
         self
     }
 
+    /// Sets the tolerance on constraints violation (cstr < tol)
     pub fn cstr_tol(&mut self, tol: f64) -> &mut Self {
         self.cstr_tol = tol;
         self
     }
 
+    /// Sets an initial DOE containing ns samples
+    ///
+    /// Either nt = nx then only x are specified and ns evals are done to get y doe values,
+    /// or nt = nx + ny then x = doe(:, :nx) and y = doe(:, nx:) are specified  
     pub fn doe(&mut self, doe: Option<Array2<f64>>) -> &mut Self {
         self.doe = doe.map(|x| x.to_owned());
         self
     }
 
+    /// Sets the parallel infill strategy
+    ///
+    /// Parallel infill criteria to get virtual next promising points in order to allow
+    /// n parallel evaluations of the function under optimization.
     pub fn qei_strategy(&mut self, q_ei: QEiStrategy) -> &mut Self {
         self.q_ei = q_ei;
         self
     }
 
+    /// Sets the nfill criteria
     pub fn infill_strategy(&mut self, infill: InfillStrategy) -> &mut Self {
         self.infill = infill;
         self
     }
 
+    /// Sets the infill optimizer
     pub fn infill_optimizer(&mut self, optimizer: InfillOptimizer) -> &mut Self {
         self.infill_optimizer = optimizer;
         self
     }
 
+    /// Sets the allowed regression models used in gaussian processes.
     pub fn regression_spec(&mut self, regression_spec: RegressionSpec) -> &mut Self {
         self.regression_spec = regression_spec;
         self
     }
 
+    /// Sets the allowed correlation models used in gaussian processes.
     pub fn correlation_spec(&mut self, correlation_spec: CorrelationSpec) -> &mut Self {
         self.correlation_spec = correlation_spec;
         self
     }
 
+    /// Sets the number of components to be used specifiying PLS projection is used (a.k.a KPLS method).
+    ///
+    /// This is used to address high-dimensional problems typically when nx > 9.
     pub fn kpls_dim(&mut self, kpls_dim: Option<usize>) -> &mut Self {
         self.kpls_dim = kpls_dim;
         self
     }
 
+    /// Sets the number of clusters used by the mixture of surrogate experts.
     pub fn n_clusters(&mut self, n_clusters: Option<usize>) -> &mut Self {
         self.n_clusters = n_clusters;
         self
     }
 
+    /// Sets a known minimum to be used as a stopping criterion.
     pub fn expect(&mut self, expected: Option<ApproxValue>) -> &mut Self {
         self.expected = expected;
         self
     }
 
+    /// Sets a directory to write optimization history and used as search path for hot start doe
     pub fn outdir(&mut self, outdir: Option<String>) -> &mut Self {
         self.outdir = outdir;
         self
     }
 
+    /// Whether we start by loading last DOE saved in `outdir` as initial DOE
     pub fn hot_start(&mut self, hot_start: bool) -> &mut Self {
         self.hot_start = hot_start;
         self
     }
 
-    pub fn moe_params(&mut self, moe_params: Option<&'a dyn MoeFit>) -> &mut Self {
-        self.moe_params = moe_params;
+    /// Sets the surrogate builder used to model objective and constraints.
+    ///
+    /// If none Mixture of Experts [MoeParams] will be used.
+    pub fn surrogate_builder(
+        &mut self,
+        surrogate_builder: Option<&'a dyn SurrogateBuilder>,
+    ) -> &mut Self {
+        self.surrogate_builder = surrogate_builder;
         self
     }
 
-    pub fn evaluator(&mut self, evaluator: Option<&'a dyn Evaluator>) -> &mut Self {
-        self.evaluator = evaluator;
+    /// Sets a pre processor for inputs before being evaluated by the function under optimization
+    ///
+    /// Used for mixed-integer optimizatio. See [crate::MixintEgor]
+    pub fn pre_proc(&mut self, pre_proc: Option<&'a dyn PreProcessor>) -> &mut Self {
+        self.pre_proc = pre_proc;
         self
     }
 
+    /// Sets a random generator for reproducibility
     pub fn with_rng<R2: Rng + Clone>(self, rng: R2) -> Egor<'a, O, R2> {
         Egor {
             n_eval: self.n_eval,
             n_start: self.n_start,
-            n_parallel: self.n_parallel,
+            q_parallel: self.q_parallel,
             n_doe: self.n_doe,
             n_cstr: self.n_cstr,
             cstr_tol: self.cstr_tol,
@@ -338,8 +386,8 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
             expected: self.expected,
             outdir: self.outdir,
             hot_start: self.hot_start,
-            moe_params: self.moe_params,
-            evaluator: self.evaluator,
+            surrogate_builder: self.surrogate_builder,
+            pre_proc: self.pre_proc,
             obj: self.obj,
             rng,
         }
@@ -421,6 +469,7 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         let mut no_point_added_retries = MAX_RETRY;
         for i in 1..=n_iter {
             let (x_dat, y_dat) = self.next_points(i, &x_data, &y_data, &sampling);
+            debug!("Try adding {}", x_dat);
             let rejected_count = update_data(&mut x_data, &mut y_data, &x_dat, &y_dat);
             if rejected_count > 0 {
                 info!(
@@ -440,7 +489,7 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
                 info!("End iteration {}/{}", i, n_iter);
             } else {
                 no_point_added_retries = MAX_RETRY;
-                let count = (self.n_parallel - rejected_count) as i32;
+                let count = (self.q_parallel - rejected_count) as i32;
                 let x_to_eval = x_data.slice(s![-count.., ..]).to_owned();
                 info!(
                     "Add {} point{}:",
@@ -478,11 +527,13 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
             }
         }
         let best_index = self.find_best_result_index(&y_data);
-        info!("{}", concatenate![Axis(1), x_data, y_data]);
-        Ok(OptimResult {
+        info!("History: \n{}", concatenate![Axis(1), x_data, y_data]);
+        let res = OptimResult {
             x_opt: x_data.row(best_index).to_owned(),
             y_opt: y_data.row(best_index).to_owned(),
-        })
+        };
+        info!("Optim Result: min f(x)={} at x= {}", res.y_opt, res.x_opt);
+        Ok(res)
     }
 
     fn next_points(
@@ -492,30 +543,31 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         y_data: &Array2<f64>,
         sampling: &Lhs<f64, R>,
     ) -> (Array2<f64>, Array2<f64>) {
+        debug!("Make surrogate with {}", x_data);
         let mut x_dat = Array2::zeros((0, x_data.ncols()));
         let mut y_dat = Array2::zeros((0, y_data.ncols()));
         let n_clusters = self.n_clusters.unwrap_or(1);
 
-        let default_params = &Moe::params(n_clusters)
-            .set_kpls_dim(self.kpls_dim)
-            .set_regression_spec(self.regression_spec)
-            .set_correlation_spec(self.correlation_spec)
-            as &dyn MoeFit;
-        let params = self.moe_params.unwrap_or(default_params);
+        let default_builder = &Moe::params(n_clusters)
+            .kpls_dim(self.kpls_dim)
+            .regression_spec(self.regression_spec)
+            .correlation_spec(self.correlation_spec)
+            as &dyn SurrogateBuilder;
+        let builder = self.surrogate_builder.unwrap_or(default_builder);
 
-        let obj_model = params
-            .fit_for_predict(x_data, &y_data.slice(s![.., 0..1]).to_owned())
+        let obj_model = builder
+            .train(x_data, &y_data.slice(s![.., 0..1]).to_owned())
             .expect("GP training failure");
 
-        let mut cstr_models: Vec<Box<dyn MoePredict>> = Vec::with_capacity(self.n_cstr);
+        let mut cstr_models: Vec<Box<dyn Surrogate>> = Vec::with_capacity(self.n_cstr);
         for k in 0..self.n_cstr {
             cstr_models.push(
-                params
-                    .fit_for_predict(x_data, &y_data.slice(s![.., k + 1..k + 2]).to_owned())
+                builder
+                    .train(x_data, &y_data.slice(s![.., k + 1..k + 2]).to_owned())
                     .expect("GP training failure"),
             )
         }
-        for _ in 0..self.n_parallel {
+        for _ in 0..self.q_parallel {
             match self.find_best_point(x_data, y_data, sampling, obj_model.as_ref(), &cstr_models) {
                 Ok(xk) => {
                     match self.get_virtual_point(&xk, y_data, obj_model.as_ref(), &cstr_models) {
@@ -549,8 +601,8 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         x_data: &ArrayBase<impl Data<Elem = f64>, Ix2>,
         y_data: &ArrayBase<impl Data<Elem = f64>, Ix2>,
         sampling: &Lhs<f64, R>,
-        obj_model: &dyn MoePredict,
-        cstr_models: &[Box<dyn MoePredict>],
+        obj_model: &dyn Surrogate,
+        cstr_models: &[Box<dyn Surrogate>],
     ) -> Result<Array1<f64>> {
         let f_min = y_data.min().unwrap();
 
@@ -578,7 +630,9 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
                         let f = |x: &Vec<f64>| -> f64 {
                             cstr_models[i]
                                 .predict_values(
-                                    &Array::from_shape_vec((1, x.len()), x.to_vec()).unwrap(),
+                                    &Array::from_shape_vec((1, x.len()), x.to_vec())
+                                        .unwrap()
+                                        .view(),
                                 )
                                 .unwrap()[[0, 0]]
                                 / params.scale_cstr[index]
@@ -586,7 +640,11 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
                         grad[..].copy_from_slice(&x.to_vec().forward_diff(&f));
                     }
                     cstr_models[index]
-                        .predict_values(&Array::from_shape_vec((1, x.len()), x.to_vec()).unwrap())
+                        .predict_values(
+                            &Array::from_shape_vec((1, x.len()), x.to_vec())
+                                .unwrap()
+                                .view(),
+                        )
                         .unwrap()[[0, 0]]
                         / params.scale_cstr[index]
                 };
@@ -599,10 +657,10 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         let mut best_x = None;
 
         let scaling_points = sampling.sample(100 * self.xlimits.nrows());
-        let scale_obj = compute_obj_scale(&scaling_points, obj_model);
-        let scale_cstr = compute_cstr_scales(&scaling_points, cstr_models);
+        let scale_obj = compute_obj_scale(&scaling_points.view(), obj_model);
+        let scale_cstr = compute_cstr_scales(&scaling_points.view(), cstr_models);
         let scale_wb2 = if self.infill == InfillStrategy::WB2S {
-            compute_wb2s_scale(&scaling_points, obj_model, *f_min)
+            compute_wb2s_scale(&scaling_points.view(), obj_model, *f_min)
         } else {
             1.
         };
@@ -689,19 +747,19 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         &self,
         xk: &ArrayBase<impl Data<Elem = f64>, Ix1>,
         y_data: &ArrayBase<impl Data<Elem = f64>, Ix2>,
-        obj_model: &dyn MoePredict,
-        cstr_models: &[Box<dyn MoePredict>],
+        obj_model: &dyn Surrogate,
+        cstr_models: &[Box<dyn Surrogate>],
     ) -> Result<Vec<f64>> {
         let mut res: Vec<f64> = Vec::with_capacity(3);
         if self.q_ei == QEiStrategy::ConstantLiarMinimum {
-            let index_min = y_data.slice(s![.., 0]).argmin().unwrap();
+            let index_min = y_data.slice(s![.., 0_usize]).argmin().unwrap();
             res.push(y_data[[index_min, 0]]);
             for ic in 1..=self.n_cstr {
                 res.push(y_data[[index_min, ic]]);
             }
             Ok(res)
         } else {
-            let x = &xk.to_owned().insert_axis(Axis(0));
+            let x = &xk.view().insert_axis(Axis(0));
             let pred = obj_model.predict_values(x)?[[0, 0]];
             let var = obj_model.predict_variances(x)?[[0, 0]];
             let conf = match self.q_ei {
@@ -721,7 +779,7 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
     fn infill_eval(
         &self,
         x: &[f64],
-        obj_model: &dyn MoePredict,
+        obj_model: &dyn Surrogate,
         f_min: f64,
         scale: f64,
         scale_wb2: f64,
@@ -734,12 +792,10 @@ impl<'a, O: GroupFunc, R: Rng + Clone> Egor<'a, O, R> {
         };
         obj / scale
     }
-}
 
-impl<'a, O: GroupFunc, R: Rng + Clone> Evaluator for Egor<'a, O, R> {
     fn eval(&self, x: &Array2<f64>) -> Array2<f64> {
-        if let Some(evaluator) = self.evaluator {
-            (self.obj)(&evaluator.eval(x).view())
+        if let Some(pre_proc) = self.pre_proc {
+            (self.obj)(&pre_proc.run(x).view())
         } else {
             (self.obj)(&x.view())
         }
@@ -768,7 +824,7 @@ mod tests {
             .infill_strategy(InfillStrategy::EI)
             .regression_spec(RegressionSpec::QUADRATIC)
             .correlation_spec(CorrelationSpec::ALL)
-            .n_eval(15)
+            .n_eval(20)
             .doe(Some(initial_doe.to_owned()))
             .expect(Some(ApproxValue {
                 value: -15.1,
@@ -778,7 +834,7 @@ mod tests {
             .minimize()
             .expect("Minimize failure");
         let expected = array![-15.1];
-        assert_abs_diff_eq!(expected, res.y_opt, epsilon = 0.3);
+        assert_abs_diff_eq!(expected, res.y_opt, epsilon = 0.5);
         let saved_doe: Array2<f64> = read_npy(DOE_INITIAL_FILE).unwrap();
         assert_abs_diff_eq!(initial_doe, saved_doe.slice(s![..3, ..1]), epsilon = 1e-6);
     }
@@ -889,10 +945,6 @@ mod tests {
             .infill_optimizer(InfillOptimizer::Cobyla) // test passes also with WB2S and Slsqp
             .doe(Some(doe))
             .n_eval(40)
-            // .expect(Some(ApproxValue {
-            //     value: -5.5080,
-            //     tolerance: 1e-3,
-            // }))
             .minimize()
             .expect("Minimize failure");
         println!("G24 optim result = {:?}", res);
